#ifndef PavementMapObject_h
#define PavementMapObject_h

// For compilers that support precompilation, includes "wx.h".
#include <wx/wxprec.h>

#ifdef __BORLANDC__
#pragma hdrstop
#endif

#ifndef WX_PRECOMP
// Include your minimal set of headers here, or wx.h
#include <wx/wx.h>
#endif

#include <Map/MapObject.h>
#include <Transactions/TransactionObjectChangeListener.h>

#include "Airport.h"

/**
 * \brief A PavementMapObject represents a rectangular area of pavement.
 *
 * The area is defined by center location, length, width and true heading.
 */
class PavementMapObject: public MapObject,
                         public TransactionObjectChangeListener {
        
        DECLARE_CLASS(PavementMapObject)
        TRANSACTIONOBJECT_DEFINE_SMARTREF(PavementMapObject);
        
        friend class PavementSelection;
        friend class PavementSelectionEndHandle;
        friend class PavementSelectionSideHandle;
public:
        typedef enum {
                Surface_Asphalt,
                Surface_Concrete,
                Surface_Turf,
                Surface_Dirt,
                Surface_Gravel,
                Surface_Helipad_Asphalt,
                Surface_Helipad_Concrete,
                Surface_Helipad_Turf,
                Surface_Helipad_Dirt,
                Surface_Twy_Asphalt_yellow_hold,
                Surface_Twy_Concrete_yellow_hold,
                Surface_Lakebed,
                Surface_Water
        } Surface;
        
        /* Create a new object */
	PavementMapObject(Airport::SmartRef airport,
                          Point3D center,
                          double length,
                          double width,
                          double heading,
                          Surface surface,
                          float smoothness);
        
        /* May only be called by MapLayer */
        virtual ~PavementMapObject();
        
        /* Paint this object using the given device context into the given editor. */
        virtual void Paint(wxDC& dc, MapEditor* editor, const MapPass& pass);
        
        /* Return TRUE iff this object intersects the given hit-TDRectangle r to list.
         * r is given in the local coordinate system of editor. */
        virtual bool Hit(MapEditor* editor, TDRectangle r);
        
        virtual wxRect GetDeviceBounds(MapEditor* editor);
        
        /* Return the geographic (lat/lon) bounds */
        virtual TDRectangle GetGeographicBounds();
        
        wxColour GetSurfaceColour() const;
        
        TRANSACTIONOBJECT_DEFINE_STATE_ACCESSOR_WCOMP(Point3D,Center,_center)
        TRANSACTIONOBJECT_DEFINE_STATE_ACCESSOR_WCOMP(float,Length,_length)
        TRANSACTIONOBJECT_DEFINE_STATE_ACCESSOR_WCOMP(float,Width,_width)
        TRANSACTIONOBJECT_DEFINE_STATE_ACCESSOR_WCOMP(float,Heading,_heading)
        TRANSACTIONOBJECT_DEFINE_STATE_ACCESSOR_WCOMP(Surface,Surface,_surface)
        TRANSACTIONOBJECT_DEFINE_STATE_ACCESSOR_WCOMP(float,Smoothness,_smoothness)
        
        virtual const Point3D* GetEndCorners() const { return _endCorners; }
        virtual const Point3D* GetEndCenters() const { return _endCenters; }
        Airport::SmartRef GetAirport() const { return _airport; }
        
        virtual void BeforeObjectChange(TransactionObject* object) {
                if (object==_airport) {
                        Damage();
                }
        }
        
        virtual void AfterObjectChange(TransactionObject* object) {
                if (object==_airport) {
                        Damage();
                        UpdateCorners();
                }
        }
        
        /**
         * \brief Pass used for rendering lighting.
         */
        static const MapPass LightingPass;
protected:
        virtual void AfterWrite() {
                UpdateCorners();
                MapObject::AfterWrite();
        }
        
        virtual void UpdateCorners();
        
        void CalculateRectCorners(double centerlat, double centerlon,
                                  double elev,
                                  double heading,
                                  double halflength[2],
                                  double width,
                                  Point3D centerEnds[2],
                                  Point3D corners[4]);
        
        Airport::SmartRef _airport;
        
        /* intermediate storage for derived positions */
        Point3D _endCenters[2];
        Point3D _endCorners[4];

        typedef struct {
                Point3D _center;
                float _length,_width;// in ft
                float _heading; // in degrees
                Surface _surface;
                float _smoothness;
        } State;
        
        TRANSACTIONOBJECT_DEFINE_STATE_MEMENTO(PavementMapObject,MapObject,State)
};

#endif
// :mode=c++-c:folding=explicit:
