#if HAVE_CONFIG_H
#  include <config.h>
#endif /* HAVE_CONFIG_H */

#include "RunwayMapObject.h"
#include "RunwaySelection.h"
#include <Map/MapEditor.h>
#include <Map/MapCoordinateSystem.h>
#include <Map/MapProjection.h>
#include <Math/sg_geodesy.h>
#include <Math/CollisionUtils.h>

IMPLEMENT_CLASS(RunwayMapObject,PavementMapObject)

RunwayMapObject::RunwayMapObject(Airport::SmartRef airport,
                                 wxString designation,
                                 Point3D center,
                                 double length,
                                 double width,
                                 double heading,
                                 Surface surface,
                                 float smoothness,
                                 RunwayEndData designatedEnd,
                                 RunwayEndData oppositeEnd,
                                 Markings markings,
                                 Shoulder shoulder,
                                 bool hasDistanceRemainingSigns):
        PavementMapObject(airport,center,length,width,heading,surface,smoothness)
{
                _myState._designation=designation;
                _myState._designatedEnd=designatedEnd;
                _myState._oppositeEnd=oppositeEnd;
                _myState._markings=markings;
                _myState._shoulder=shoulder;
                _myState._hasDistanceRemainingSigns=hasDistanceRemainingSigns;
                UpdateCorners();
}

RunwayMapObject::~RunwayMapObject() {
}

void RunwayMapObject::Paint(wxDC& dc, MapEditor* editor, const MapPass& pass) {
        if (pass!=StandardPass)
                return;
        wxPoint corners[4];
        wxPoint localDispThreshCorners[4];
        wxPoint localStopwayCorners[4];
        MapProjection* proj=editor->GetProjection();
        MapCoordinateSystem* cs=&editor->GetCoordinateSystem();
        for (int i=0;i<4;i++) {
                corners[i]=cs->LogicalToDevice(proj->ConvertToLocal(_endCorners[i]));
                localDispThreshCorners[i]=cs->LogicalToDevice(proj->ConvertToLocal(_dispThreshCorners[i]));
                localStopwayCorners[i]=cs->LogicalToDevice(proj->ConvertToLocal(_stopwayCorners[i]));
        }

        dc.SetPen(*wxWHITE_PEN);
        dc.SetBrush(wxBrush(GetSurfaceColour()));
        dc.DrawPolygon(4,localStopwayCorners);
        dc.SetBrush(wxNullBrush);
        dc.SetPen(wxNullPen);

        wxPen stopwayPen(wxColour(190,190,0));
        wxBrush stopwayBrush(wxColour(190,190,0));
        
        // Draw stopways
        dc.SetPen(stopwayPen);
        dc.SetBrush(stopwayBrush);
        
        wxPoint stopwayPoints[4];
        if (GetDesignatedEnd().stopwayLength>0) {
                stopwayPoints[0]=corners[0];
                stopwayPoints[1]=corners[1];
                stopwayPoints[2]=localStopwayCorners[1];
                stopwayPoints[3]=localStopwayCorners[0];
                dc.DrawPolygon(4,stopwayPoints);
        }
        if (GetOppositeEnd().stopwayLength>0) {
                stopwayPoints[0]=corners[2];
                stopwayPoints[1]=corners[3];
                stopwayPoints[2]=localStopwayCorners[3];
                stopwayPoints[3]=localStopwayCorners[2];
                dc.DrawPolygon(4,stopwayPoints);
        }
        
        // Draw thick lines for threshold
        wxPen threshPen(*wxWHITE,3);
        dc.SetPen(threshPen);

        dc.DrawLine(localDispThreshCorners[0].x,localDispThreshCorners[0].y,
                    localDispThreshCorners[1].x,localDispThreshCorners[1].y);
        dc.DrawLine(localDispThreshCorners[2].x,localDispThreshCorners[2].y,
                    localDispThreshCorners[3].x,localDispThreshCorners[3].y);
        
        dc.SetBrush(wxNullBrush);
        dc.SetPen(wxNullPen);
}

bool RunwayMapObject::Hit(MapEditor* editor, TDRectangle r) {
        TDRectangle bounds=GetLocalBounds(editor);
        if (!bounds.Intersects(r))
                return false;
        MapProjection* proj=editor->GetProjection();
        Point3D corners[4];
        for (unsigned int i=0;i<4;i++)
                corners[i]=proj->ConvertToLocal(_stopwayCorners[i]);
        return CollisionUtils::ConvexPolygonRectangleCollide(r, 4, corners);
}

MapSelection* RunwayMapObject::CreateSelection(MapSelectionManager* manager) {
        return new RunwaySelection(manager,this);
}

wxRect RunwayMapObject::GetDeviceBounds(MapEditor* editor) {
        wxCoord top,bottom,left,right;
        
        MapProjection* proj=editor->GetProjection();
        MapCoordinateSystem* cs=&editor->GetCoordinateSystem();
        
        wxPoint corner;
        corner=cs->LogicalToDevice(proj->ConvertToLocal(_stopwayCorners[0]));
        left=right=corner.x;
        top=bottom=corner.y;
        
        for (int i=1;i<4;i++) {
                corner=cs->LogicalToDevice(proj->ConvertToLocal(_stopwayCorners[i]));
                left=std::min(left,corner.x);
                right=std::max(right,corner.x);
                top=std::min(top,corner.y);
                bottom=std::max(bottom,corner.y);
        }
        return wxRect(left,top,right-left,bottom-top);
}

/* Return the geographic bounds (lat/lon) */
TDRectangle RunwayMapObject::GetGeographicBounds() {
        TDRectangle rect;
        for (int i=0;i<4;i++)
                rect=TDRectangle::Union(rect,TDRectangle(_stopwayCorners[i]));
        return rect;
}

wxString RunwayMapObject::GetOppositeDesignation() const {
        wxString number=_myState._designation.Left(2);
        wxString differentiator=_myState._designation.Right(1);
        
        long intNumber;
        number.ToLong(&intNumber);
        if (intNumber>18)
                intNumber-=18;
        else
                intNumber+=18;
        
        if (differentiator.CmpNoCase(wxT("L"))==0)
                differentiator=wxT("R");
        else if (differentiator.CmpNoCase(wxT("R"))==0)
                differentiator=wxT("L");
        
        return wxString::Format(wxT("%02d%s"),intNumber,differentiator.c_str());
}

void RunwayMapObject::UpdateCorners() {
        PavementMapObject::UpdateCorners();
        double halflength[2];

        RunwayEndData designatedEnd,oppositeEnd;
        
        designatedEnd=GetDesignatedEnd();
        oppositeEnd=GetOppositeEnd();
        
        halflength[0]=(GetLength()/2.0-designatedEnd.displacedThreshold)*0.3048;
        halflength[1]=(GetLength()/2.0-oppositeEnd.displacedThreshold)*0.3048;
        
        double alt=GetAirport()->GetElevation()*0.3048;
        
        CalculateRectCorners(GetCenter().lat(),GetCenter().lon(),
                             alt,
                             GetHeading(),
                             halflength,
                             GetWidth()*0.3048,
                             _dispThreshCenterEnds,
                             _dispThreshCorners);

        halflength[0]=(GetLength()/2.0+designatedEnd.stopwayLength)*0.3048;
        halflength[1]=(GetLength()/2.0+oppositeEnd.stopwayLength)*0.3048;
        
        CalculateRectCorners(GetCenter().lat(),GetCenter().lon(),
                             alt,
                             GetHeading(),
                             halflength,
                             GetWidth()*0.3048,
                             _stopwayCenterEnds,
                             _stopwayCorners);
}
