#if HAVE_CONFIG_H
#  include <config.h>
#endif /* HAVE_CONFIG_H */

#include "OldProjectLoader.h"

#include "XPlaneLoader.h"

#include <AI/AIGroundNetworkLoader.h>
#include <Imagery/MapCalibratedBackgroundImage.h>
#include <Map/MapObjectContainer.h>
#include <Transactions/TransactionManager.h>

#include <xml/easyxml.hxx>

#if _MSC_VER >= 1400
#  define strtok_r strtok_s
#elif _MSC_VER >= 1300
#  define strtok_r(a,b,c) strtok(a,b)
#endif

#include <fstream>

IMPLEMENT_CLASS(OldProjectLoader,wxObject);

OldProjectLoader::OldProjectLoader(TransactionManager* tmgr,
                                   std::istream& infile):
           _transactionManager(tmgr),
           _infile(infile)
{
}

OldProjectLoader::~OldProjectLoader() {
}

Airport::SmartRef OldProjectLoader::LoadProject(wxProgressDialog* progdlg,
                                                wxString& projection,
                                                MapObjectContainer::SmartRef backgroundImageLayer) {
        char buf[2000],*token,*saveptr;
        const char space[]=" \n\t\r";
        wxString groundNetPath;
        int xplaneVersion=0;
        
        Airport::SmartRef theAirport;
        
        while (!_infile.getline(buf,sizeof(buf)).fail()) {
                token = strtok_r(buf, space, &saveptr);
                if(!strcmp(token, "[FORMAT]")) {
                        if(!_infile.getline(buf,sizeof(buf)).fail()) {
                                token = strtok_r(buf, space, &saveptr);
                                if(!strcmp(token, "FGFS")) {
                                        // FGFS format is currently not supported
                                        return NULL;
                                } else if (!strcmp(token, "X-PLANE")) {
                                        // try to identify the version marker
                                        token=strtok_r(NULL, space, &saveptr);
                                        if (token!=NULL) {
                                                xplaneVersion=atoi(token);
                                        }
                                }
                        }
                } else if (!strcmp(token,"[CALIBRATION]")) {
                        LoadCalibratedBackgroundImage(progdlg,projection,backgroundImageLayer);
                } else if (!strcmp(token,"[GROUNDNET]")) {
                        LoadGroundNetwork(groundNetPath);
                } else if (!strcmp(token,"[AIRPORT]")) {
                        if (_infile.getline(buf,sizeof(buf)).fail())
                                return NULL;
                        XPlaneLoader xploader(_transactionManager,_infile);
                        xploader.setVersion(xplaneVersion);
                        progdlg->Update(-1,wxT("Loading airport data"));
                        theAirport=xploader.scanAirport(buf);
                        break;
                }
        }
        
        if (!theAirport)
                return theAirport;
        
        if (!groundNetPath.IsEmpty()) {
                progdlg->Update(-1,wxT("Loading Ground Network"));
                
                string cpath=(const char*)groundNetPath.mb_str(*wxConvFileName);
                
                std::ifstream istream(cpath.c_str());
                
                AIGroundNetworkLoader aiNetworkLoader(theAirport);
                
                readXML(istream,aiNetworkLoader,cpath);
                
                istream.close();
        }
        
        return theAirport;
}

void OldProjectLoader::LoadCalibratedBackgroundImage(wxProgressDialog* progdlg,
                                                     wxString& projection,
                                                     MapObjectContainer::SmartRef backgroundImageLayer) {
        char buf[2000],*token,*saveptr;
        const char space[]=" \n\t\r";
        const char junk[]="?\n\r";
        
        bool scaleFound=false,
                angleFound=false,
                latFound=false,
                lonFound=false,
                fileFound=false,
                projectionFound=false;
        float scale,angle=0.0,lat,lon;
        wxString file,imageProjection;
        
        while (!_infile.getline(buf,sizeof(buf)).fail()) {
                token = strtok_r(buf, space, &saveptr);
                if (!strcmp(token,"[END-CALIBRATION]")) {
                        break;
                } else if (!strcmp(token,"SCALE")) {
                        token = strtok_r(NULL, space, &saveptr);
                        scale=atof(token);
                        scaleFound=true;
                } else if (!strcmp(token,"ANGLE")) {
                        token = strtok_r(NULL, space, &saveptr);
                        angle=atof(token);
                        angleFound=true;
                } else if (!strcmp(token,"LAT")) {
                        token = strtok_r(NULL, space, &saveptr);
                        lat=atof(token);
                        latFound=true;
                } else if (!strcmp(token,"LON")) {
                        token = strtok_r(NULL, space, &saveptr);
                        lon=atof(token);
                        lonFound=true;
                } else if (!strcmp(token,"FILE")) {
                        token = strtok_r(NULL, junk, &saveptr);
                        file=wxString(token,wxConvUTF8);
                        fileFound=true;
                } else if (!strcmp(token,"PROJECTION")) {
                        token = strtok_r(NULL, junk, &saveptr);
                        imageProjection=wxString(token,wxConvUTF8);
                        projectionFound=true;
                }
        }
        
        if (projectionFound && !projection.IsEmpty() && projection.Cmp(imageProjection)) {
                ::wxMessageBox(wxT("Projection mismatch on loading multiple ")
                               wxT("calibrated images from project file.\n")
                               wxT("Will use projection ")+projection+wxT("."),
                               wxT("Warning"),
                               wxICON_INFORMATION|wxOK);
        }
        
        if (!fileFound || !latFound || !lonFound || !scaleFound) {
                ::wxMessageBox(wxT("Incomplete calibration found for image"),
                               wxT("Error"),
                               wxICON_ERROR|wxOK);
                return;
        }
        
        projection=imageProjection;
        
        if (!::wxFileExists(file)) {
                ::wxMessageBox(wxT("Source file for calibrated image does not exist"),
                               wxT("Error"),
                               wxICON_ERROR|wxOK);
                return;
        }
        
        progdlg->Update(-1,wxT("Loading background image"));
        
        MapCalibratedBackgroundImage::SmartRef imageObject;
        Point3D northWestPos;
        northWestPos.setlat(lat);
        northWestPos.setlon(lon);
        imageObject=new MapCalibratedBackgroundImage(_transactionManager,
                                        northWestPos,
                                        scale,
                                        angle);
        imageObject->SetFilename(file);
        imageObject->LoadImage();
        backgroundImageLayer->AttachObject(imageObject);
        
        // make sure that all preparations are done
        imageObject->GetImage();
        
        imageObject->Damage();
}

void OldProjectLoader::LoadGroundNetwork(wxString& groundNetPath) {
        char buf[2000],*token,*saveptr;
        const char space[]=" \n\t\r";
        const char junk[]="?\n\r";
        
        while (!_infile.getline(buf,sizeof(buf)).fail()) {
                token = strtok_r(buf, space, &saveptr);
                if (!strcmp(token,"[END-GROUNDNET]")) {
                        break;
                } else if (!strcmp(token,"FILE")) {
                        token = strtok_r(NULL, junk, &saveptr);
                        groundNetPath=wxString(token,wxConvUTF8);
                }
        }
}
