#if HAVE_CONFIG_H
#  include <config.h>
#endif /* HAVE_CONFIG_H */

#include "XPlaneLoader.h"

#include <Airport/BeaconMapObject.h>
#include <Airport/PavementMapObject.h>
#include <Airport/RunwayMapObject.h>
#include <Airport/TaxiwayMapObject.h>
#include <Airport/TowerMapObject.h>
#include <Airport/WindsockMapObject.h>

#include <AI/StartupLocationMapObject.h>

#if _MSC_VER >= 1400
#  define strtok_r strtok_s
#elif _MSC_VER >= 1300
#  define strtok_r(a,b,c) strtok(a,b)
#endif
#ifdef _MSC_VER
#  define strcasecmp _stricmp
#endif

IMPLEMENT_CLASS(XPlaneLoader,wxObject);

XPlaneLoader::XPlaneLoader(TransactionManager* tmgr, 
                           std::istream& infile):
           _transactionManager(tmgr),
           _infile(&infile),
           _version(0) {
                   
}

XPlaneLoader::~XPlaneLoader() {
}

Airport::SmartRef XPlaneLoader::loadAirport(const std::string& icao) {
        char buf[2000],*token,*saveptr;
        char linestr[2000];
        
        const char space[]=" \n\t\r";
        
        while (!_infile->getline(buf,sizeof(buf)).fail()) {
                strcpy(linestr,buf);
                token=strtok_r(buf,space,&saveptr);
                if (token==NULL)
                        continue; // skip blank lines
                if (!strcmp(token,"99"))
                        break; // last line
                
                if (!strcmp(token,"715")) {
                        _version=715;
                        continue;
                } else if (!strcmp(token,"810")) {
                        _version=810;
                        continue;
                } else if (!strcmp(token,"840")) {
                        _version=840;
                        continue;
                } else if (!strcmp(token,"1") ||
                           !strcmp(token,"16") ||
                           !strcmp(token,"17")) {
                        /* Found an airport line, skip to ICAO ID */
                        if (!strtok_r(NULL,space,&saveptr) ||
                            !strtok_r(NULL,space,&saveptr) ||
                            !strtok_r(NULL,space,&saveptr) ||
                            !(token=strtok_r(NULL,space,&saveptr)))
                                continue;
                        if (icao==token) {
                                return scanAirport(linestr);
                        }
                }
        }
        return Airport::SmartRef();
}

Airport::SmartRef XPlaneLoader::scanAirport(char* header) {
        Airport::SmartRef airport;
        Airport::AirportType apt_type;
        float elevation;
        bool hasControlTower;
        bool hasDefaultBuildings;
        wxString icaoID;
        wxString name;
        FrequencyEntryArray frequencies;
        
        char buf[2000],*token,*saveptr;
        char linestr[2000];
        
        const char space[]=" \n\t\r";
        const char junk[]="?\n\r";
        
        // airport type
        token=strtok_r(header,space,&saveptr);
        wxASSERT_MSG(token!=NULL,wxT("scanAirport called with an empty header string"));
        if (!strcmp(token,"1")) {
                apt_type=Airport::AirportType_Land;
        } else if (!strcmp(token,"16")) {
                apt_type=Airport::AirportType_Sea;
        } else if (!strcmp(token,"17")) {
                apt_type=Airport::AirportType_Heliport;
        } else
                return NULL;
        
        // elevation
        token=strtok_r(NULL,space,&saveptr);
        if (!token)
                return NULL;
        elevation=atof(token);
        
        // "has control tower" flag
        token=strtok_r(NULL,space,&saveptr);
        if (!token)
                return NULL;
        hasControlTower=(atoi(token)!=0);
        
        // "has default buildings" flag
        token=strtok_r(NULL,space,&saveptr);
        if (!token)
                return NULL;
        hasDefaultBuildings=(atoi(token)!=0);
        
        // ICAO id
        token=strtok_r(NULL,space,&saveptr);
        if (!token)
                return NULL;
        icaoID=wxString(token,wxConvUTF8);

        // name
        token=strtok_r(NULL,junk,&saveptr);
        if (!token)
                return NULL;
        name=wxString(token,wxConvUTF8);
        
        airport=new Airport(_transactionManager,
                            icaoID,name,
                            apt_type,
                            elevation,
                            hasControlTower,hasDefaultBuildings);
        
        //std::cout << "Reading airport " << header << "\n";
        while (!_infile->getline(buf,sizeof(buf)).fail()) {
                strcpy(linestr,buf);
                token=strtok_r(buf,space,&saveptr);
                if (token==NULL)
                        continue; // skip blank lines
                //std::cout << "token=" << token << "\n";
                if (!strcmp(token,"99"))
                        break; // last line
                
                if (!strcmp(token,"1") ||
                    !strcmp(token,"16") ||
                    !strcmp(token,"17")) {
                        /* airport line for the next airport */
                        break;
                } else if (!strcmp(token,"10")) {
                        /* runway or taxiway */
                        scanPavementLine(airport,linestr);
                } else if (!strcmp(token,"14")) {
                        scanTowerLocationLine(airport,linestr);
                } else if (!strcmp(token,"15")) {
                        scanStartupLocationLine(airport,linestr);
                } else if (!strcmp(token,"18")) {
                        scanBeaconLine(airport,linestr);
                } else if (!strcmp(token,"19")) {
                        scanWindsockLine(airport,linestr);
                } else if (atoi(token)>=50 && atoi(token)<60) {
                        scanFrequencyLine(airport,linestr,frequencies);
                }
        }
        
        airport->SetFrequencies(frequencies);

        frequencies=airport->GetFrequencies();
        
        return airport;
}

void XPlaneLoader::scanTowerLocationLine(Airport::SmartRef airport,char* buf) {
        char *token,*saveptr;
        Point3D loc;
        float height;
        bool drawTower=true;
        wxString name;
        
        const char space[]=" \n\t\r";
        const char junk[]="?\n\r";

        token=strtok_r(buf,space,&saveptr);
        token=strtok_r(NULL,space,&saveptr);
        if (!token)
                return;
        loc.setlat(atof(token));
        
        token=strtok_r(NULL,space,&saveptr);
        if (!token)
                return;
        loc.setlon(atof(token));
        
        // height
        token=strtok_r(NULL,space,&saveptr);
        if (!token)
                return;
        height=atof(token);
        
        if (_version>=715) {
                // "draw tower" flag (added in version 715)
                token=strtok_r(NULL,space,&saveptr);
                if (!token)
                        return;
                drawTower=(atoi(token)!=0);
        }
        
        // name
        token=strtok_r(NULL,junk,&saveptr);
        if (!token)
                return;
        name=wxString(token,wxConvUTF8);
        
        TowerMapObject::SmartRef tower;
        tower=new TowerMapObject(airport->GetTransactionManager(),loc,height,drawTower,name);
        /* X-Plane stores the objects in top-to-bottom order,
         * so we need to add all objects at the bottom */
        airport->GetObjectsLayer()->AttachObject(tower, 0);
}

void XPlaneLoader::scanStartupLocationLine(Airport::SmartRef airport,char* buf) {
        char *token,*saveptr;
        Point3D loc;
        float heading;
        wxString name;
        
        const char space[]=" \n\t\r";
        const char junk[]="?\n\r";

        token=strtok_r(buf,space,&saveptr);
        token=strtok_r(NULL,space,&saveptr);
        if (!token)
                return;
        loc.setlat(atof(token));
        
        token=strtok_r(NULL,space,&saveptr);
        if (!token)
                return;
        loc.setlon(atof(token));
        
        // heading
        token=strtok_r(NULL,space,&saveptr);
        if (!token)
                return;
        heading=atof(token);
        
        // name
        token=strtok_r(NULL,junk,&saveptr);
        if (!token)
                return;
        name=wxString(token,wxConvUTF8);

        StartupLocationMapObject::SmartRef startupLocation;
        startupLocation=new StartupLocationMapObject(airport,loc,heading,name);
        /* X-Plane stores the objects in top-to-bottom order,
         * so we need to add all objects at the bottom */
        airport->GetGroundNetworkLayer()->AttachObject(startupLocation, 0);
}

void XPlaneLoader::scanBeaconLine(Airport::SmartRef airport,char* buf) {
        char *token,*saveptr;
        Point3D loc;
        BeaconMapObject::BeaconType beaconType;
        wxString name;
        
        const char space[]=" \n\t\r";
        const char junk[]="?\n\r";

        token=strtok_r(buf,space,&saveptr);
        token=strtok_r(NULL,space,&saveptr);
        if (!token)
                return;
        loc.setlat(atof(token));
        token=strtok_r(NULL,space,&saveptr);
        if (!token)
                return;
        loc.setlon(atof(token));

        // beacon type
        token=strtok_r(NULL,space,&saveptr);
        if (!token)
                return;
        switch (atoi(token)) {
        case 1:
                beaconType=BeaconMapObject::Beacon_WhiteGreen_Flash;
                break;
        case 2:
                beaconType=BeaconMapObject::Beacon_WhiteYellow_Flash;
                break;
        case 3:
                beaconType=BeaconMapObject::Beacon_GreenYellowWhite_Flash;
                break;
        case 4:
                beaconType=BeaconMapObject::Beacon_WhiteWhiteGreen_Flash;
                break;
        case 5:
                beaconType=BeaconMapObject::Beacon_White_Strobe;
                break;
        default:
        case 0:
                beaconType=BeaconMapObject::Beacon_None;
                break;
        }
        
        // name
        token=strtok_r(NULL,junk,&saveptr);
        if (!token)
                return;
        name=wxString(token,wxConvUTF8);
        
        BeaconMapObject::SmartRef beacon;
        beacon=new BeaconMapObject(airport->GetTransactionManager(),loc,beaconType,name);
        /* X-Plane stores the objects in top-to-bottom order,
         * so we need to add all objects at the bottom */
        airport->GetObjectsLayer()->AttachObject(beacon, 0);
}

void XPlaneLoader::scanWindsockLine(Airport::SmartRef airport,char* buf) {
        char *token,*saveptr;
        Point3D loc;
        bool isLighted;
        wxString name;
        
        const char space[]=" \n\t\r";
        const char junk[]="?\n\r";

        token=strtok_r(buf,space,&saveptr);
        token=strtok_r(NULL,space,&saveptr);
        if (!token)
                return;
        loc.setlat(atof(token));
        token=strtok_r(NULL,space,&saveptr);
        if (!token)
                return;
        loc.setlon(atof(token));
        
        // "is lighted" flag
        token=strtok_r(NULL,space,&saveptr);
        if (!token)
                return;
        isLighted=(atoi(token)!=0);
        
        // name
        token=strtok_r(NULL,junk,&saveptr);
        if (!token)
                return;
        name=wxString(token,wxConvUTF8);

        WindsockMapObject::SmartRef windsock;
        windsock=new WindsockMapObject(airport->GetTransactionManager(),loc,isLighted,name);
        /* X-Plane stores the objects in top-to-bottom order,
         * so we need to add all objects at the bottom */
        airport->GetObjectsLayer()->AttachObject(windsock, 0);
}

void XPlaneLoader::scanPavementLine(Airport::SmartRef airport,char* buf) {
        char *token,*saveptr;
        char *marker;
        Point3D center;
        double length;
        double width;
        double heading;
        
        RunwayMapObject::Surface surface;
        float smoothness;
        
        wxString designation;
        RunwayMapObject::RunwayEndData designatedEnd;
        RunwayMapObject::RunwayEndData oppositeEnd;
        RunwayMapObject::Markings markings;
        RunwayMapObject::Shoulder shoulder;
        bool hasDistanceRemainingSigns;
        bool hasEdgeLights,hasCenterLights;
        
        bool isTaxiway=false;
        
        const char space[]=" \n\t\r";
        
        designatedEnd.vasiAngle=3.0;
        oppositeEnd.vasiAngle=3.0;

        token=strtok_r(buf,space,&saveptr);

        // latitude
        token=strtok_r(NULL,space,&saveptr);
        if (!token)
                return;
        center.setlat(atof(token));

        // longitude
        token=strtok_r(NULL,space,&saveptr);
        if (!token)
                return;
        center.setlon(atof(token));

        // rwy or taxiway number
        token=strtok_r(NULL,space,&saveptr);
        if (!token)
                return;
        if (!strcasecmp(token,"xxx"))
                isTaxiway=true;
        else
                designation=wxString(token,wxConvUTF8);

        // true heading
        token=strtok_r(NULL,space,&saveptr);
        if (!token)
                return;
        heading=atof(token);
        
        // length in feet
        token=strtok_r(NULL,space,&saveptr);
        if (!token)
                return;
        length=atof(token);
        
        // threshold length
        token=strtok_r(NULL,space,&saveptr);
        if (!token)
                return;
        marker=strchr(token,'.');
        if (!marker)
                return;
        *marker=0;
        designatedEnd.displacedThreshold=atoi(token);
        oppositeEnd.displacedThreshold=atoi(marker+1);
        
        // stopway length
        token=strtok_r(NULL,space,&saveptr);
        if (!token)
                return;
        marker=strchr(token,'.');
        if (!marker)
                return;
        *marker=0;
        designatedEnd.stopwayLength=atoi(token);
        oppositeEnd.stopwayLength=atoi(marker+1);
        
        // width in feet
        token=strtok_r(NULL,space,&saveptr);
        if (!token)
                return;
        width=atof(token);
        
        // lighting codes
        token=strtok_r(NULL,space,&saveptr);
        if (!token || strlen(token)!=6)
                return;
        designatedEnd.visualApproachPathLighting=GetVAPLFromCode(token[0]-'0');
        designatedEnd.runwayLighting=GetRWYLFromCode(token[1]-'0');
        designatedEnd.approachLighting=GetAPPLFromCode(token[2]-'0');
        oppositeEnd.visualApproachPathLighting=GetVAPLFromCode(token[3]-'0');
        oppositeEnd.runwayLighting=GetRWYLFromCode(token[4]-'0');
        oppositeEnd.approachLighting=GetAPPLFromCode(token[5]-'0');
        
        // setup lighting information for taxiways
        hasEdgeLights=false;
        hasCenterLights=false;
        
        /* We actually lose information here, but the X-Plane specification
         * doesn't give us any other choice.
         * The 4-code is there to be backwards compatible.
         */
        if (token[4]=='6' || token[4]=='4') {
                hasCenterLights=true;
                hasEdgeLights=true;
        }
        
        // surface codes
        token=strtok_r(NULL,space,&saveptr);
        if (!token)
                return;
        surface=GetSurfaceFromCode(atoi(token));
        
        // shoulder codes
        token=strtok_r(NULL,space,&saveptr);
        if (!token)
                return;
        shoulder=GetShoulderFromCode(atoi(token));

        // marking codes
        token=strtok_r(NULL,space,&saveptr);
        if (!token)
                return;
        markings=GetMarkingsFromCode(atoi(token));

        // smoothness
        token=strtok_r(NULL,space,&saveptr);
        if (!token)
                return;
        smoothness=atof(token);

        // distance remaining sign flag
        token=strtok_r(NULL,space,&saveptr);
        if (!token)
                return;
        hasDistanceRemainingSigns=!!atoi(token);
        
        // VASI angle (only for runways; added in version 810)
        if (!isTaxiway && _version>=810) {
                token=strtok_r(NULL,space,&saveptr);
                if (!token)
                        return;
                marker=strchr(token,'.');
                if (!marker)
                        return;
                *marker=0;
                designatedEnd.vasiAngle=atoi(token)/100.0;
                oppositeEnd.vasiAngle=atoi(marker+1)/100.0;
        }

        /* X-Plane stores the objects in top-to-bottom order,
         * so we need to add all objects at the bottom */
        if (isTaxiway) {
                TaxiwayMapObject::SmartRef taxiway;
                taxiway=new TaxiwayMapObject(airport,
                                             center,length,width,heading,
                                             surface,smoothness,
                                             hasEdgeLights,
                                             hasCenterLights);
                airport->GetTaxiwayLayer()->AttachObject(taxiway, 0);
        } else {
                RunwayMapObject::SmartRef runway;
                runway=new RunwayMapObject(airport,
                                           designation,
                                           center,length,width,heading,
                                           surface,smoothness,
                                           designatedEnd,
                                           oppositeEnd,
                                           markings,
                                           shoulder,
                                           hasDistanceRemainingSigns);
                airport->GetRunwayLayer()->AttachObject(runway, 0);
        }

}

void XPlaneLoader::scanFrequencyLine(Airport::SmartRef airport,char* buf, FrequencyEntryArray& entries) {
        char *token,*saveptr;
        FrequencyEntry::FrequencyType type;
        wxString frequency;
        wxString name;
        
        const char space[]=" \n\t\r";
        const char junk[]="?\n\r";

        // code
        token=strtok_r(buf,space,&saveptr);
        if (!token)
                return;
        switch (atoi(token)) {
        case 50:
                type=FrequencyEntry::FrequencyType_AWOS;
                break;
        case 51:
                type=FrequencyEntry::FrequencyType_UNICOM;
                break;
        case 52:
                type=FrequencyEntry::FrequencyType_Clearance;
                break;
        case 53:
                type=FrequencyEntry::FrequencyType_Ground;
                break;
        case 54:
                type=FrequencyEntry::FrequencyType_Tower;
                break;
        case 55:
                type=FrequencyEntry::FrequencyType_Approach;
                break;
        case 56:
                type=FrequencyEntry::FrequencyType_Departure;
                break;
        default:
                std::cerr << "Unknown frequency type " << token << "\n";
                return;
        }

        // frequency
        token=strtok_r(NULL,space,&saveptr);
        if (!token)
                return;
        frequency=wxString(token,wxConvUTF8);
        
        // name
        token=strtok_r(NULL,junk,&saveptr);
        if (!token)
                return;
        name=wxString(token,wxConvUTF8);
        
        entries.Add(FrequencyEntry(type,frequency,name));
}

RunwayMapObject::VisualApproachPathLighting XPlaneLoader::GetVAPLFromCode(int code) {
        switch (code) {
        case 1:
                return RunwayMapObject::VAPL_None;
        case 2:
                return RunwayMapObject::VAPL_VASI;
        case 3:
                return RunwayMapObject::VAPL_PAPI;
        case 4:
                return RunwayMapObject::VAPL_SpaceShuttlePAPI;
        }
        return RunwayMapObject::VAPL_None;
}

RunwayMapObject::RunwayLighting XPlaneLoader::GetRWYLFromCode(int code) {
        switch (code) {
        case 1:
                return RunwayMapObject::RWYL_None;
        case 2:
                return RunwayMapObject::RWYL_EdgeLights;
        case 3:
                return RunwayMapObject::RWYL_REIL;
        case 4:
                return RunwayMapObject::RWYL_CLL;
        case 5:
                return RunwayMapObject::RWYL_TDZ;
        }
        return RunwayMapObject::RWYL_None;
}

RunwayMapObject::ApproachLighting XPlaneLoader::GetAPPLFromCode(int code) {
        switch (code) {
        case 1:
                return RunwayMapObject::APPL_None;
        case 2:
                return RunwayMapObject::APPL_SSALS;
        case 3:
                return RunwayMapObject::APPL_SALSF;
        case 4:
                return RunwayMapObject::APPL_ALSF_I;
        case 5:
                return RunwayMapObject::APPL_ALSF_II;
        case 6:
                return RunwayMapObject::APPL_ODALS;
        case 7:
                return RunwayMapObject::APPL_Calvert_1;
        case 8:
                return RunwayMapObject::APPL_Calvert_2_3;
        }
        return RunwayMapObject::APPL_None;
}

PavementMapObject::Surface XPlaneLoader::GetSurfaceFromCode(int code) {
        switch (code) {
        case 1:
                return PavementMapObject::Surface_Asphalt;
        case 2:
                return PavementMapObject::Surface_Concrete;
        case 3:
                return PavementMapObject::Surface_Turf;
        case 4:
                return PavementMapObject::Surface_Dirt;
        case 5:
                return PavementMapObject::Surface_Gravel;
        case 6:
                return PavementMapObject::Surface_Helipad_Asphalt;
        case 7:
                return PavementMapObject::Surface_Helipad_Concrete;
        case 8:
                return PavementMapObject::Surface_Helipad_Turf;
        case 9:
                return PavementMapObject::Surface_Helipad_Dirt;
        case 10:
                return PavementMapObject::Surface_Twy_Asphalt_yellow_hold;
        case 11:
                return PavementMapObject::Surface_Twy_Concrete_yellow_hold;
        case 12:
                return PavementMapObject::Surface_Lakebed;
        case 13:
                return PavementMapObject::Surface_Water;
        }
        return PavementMapObject::Surface_Asphalt;
}

RunwayMapObject::Shoulder XPlaneLoader::GetShoulderFromCode(int code) {
        switch (code) {
        case 0:
                return RunwayMapObject::Shoulder_None;
        case 1:
                return RunwayMapObject::Shoulder_Asphalt;
        case 2:
                return RunwayMapObject::Shoulder_Concrete;
        }
        return RunwayMapObject::Shoulder_None;
}

RunwayMapObject::Markings XPlaneLoader::GetMarkingsFromCode(int code) {
        switch (code) {
        case 0:
                return RunwayMapObject::Markings_None;
        case 1:
                return RunwayMapObject::Markings_Visual;
        case 2:
                return RunwayMapObject::Markings_NonPrecision;
        case 3:
                return RunwayMapObject::Markings_Precision;
        }
        return RunwayMapObject::Markings_None;
}


