#if HAVE_CONFIG_H
#  include <config.h>
#endif /* HAVE_CONFIG_H */

#include "MapEditor.h"
#include "MapGrid.h"
#include "MapLayer.h"
#include <Map/MapProjection.h>

#include <wx/listimpl.cpp>
#include <wx/menuitem.h>

WX_DEFINE_LIST(MapLayerInstanceList);
WX_DEFINE_LIST(MapPassInstanceList);
WX_DEFINE_LIST(MapEditorList);

IMPLEMENT_CLASS(MapEditor,wxEvtHandler)

BEGIN_EVENT_TABLE(MapEditor, wxEvtHandler)
        EVT_PAINT(MapEditor::OnPaint)
        EVT_SIZE(MapEditor::OnResize)
        EVT_ENTER_WINDOW(MapEditor::OnMouseEnter)
        EVT_SCROLLWIN(MapEditor::OnScrollWin)
END_EVENT_TABLE()

float MapEditor::Zoom_Step_Factor=1.1;
float MapEditor::Scroll_Step_Fraction=0.1;
float MapEditor::Scroll_Line_Width=10.0;
float MapEditor::Scroll_Line_Height=10.0;
wxColour MapEditor::Map_Background_Colour=wxColour(255,255,255);
MapEditorList MapEditor::_editorList;

MapEditor::MapEditor(wxWindow* drawingWindow,
                     MapProjection* projection,
                     TransactionManager* tmgr):
        _transactionManager(tmgr),
        _selectionManager(new MapSelectionManager(this)),
        _modeManager(this),
        _projection(projection),
        _drawingWindow(drawingWindow)
{
        _transactionManager->Hold();
        _mapPassInstances.DeleteContents(true);

        _drawingWindow->PushEventHandler(&_modeManager);
        _drawingWindow->PushEventHandler(this);

        _grid=new MapGrid(this);
        _foregroundDecos.Append(_grid);
        _editorList.Append(this);
}

MapEditor::~MapEditor() {
        _editorList.DeleteObject(this);
        _drawingWindow->RemoveEventHandler(this);
        _drawingWindow->RemoveEventHandler(&_modeManager);

        MapLayerInstanceList::Node *layerNode;
        while ((layerNode=_mapLayerInstances.GetFirst())!=NULL) {
                layerNode->GetData()->GetLayer()->UnregisterEditor(this);
                delete layerNode->GetData();
                _mapLayerInstances.DeleteNode(layerNode);
        }

        if (_projection)
                delete _projection;
        
        _transactionManager->Unhold();
}

MapProjection* MapEditor::SetProjection(MapProjection* projection) {
        MapProjection* oldproj=_projection;
        wxASSERT(_projection!=NULL);
        
        /* get the old geographical bounds */
        TDRectangle geoBounds=GetVisibleGeographicBounds();
        
        _projection=projection;
        
        /* reset the canvas local bounds */
        _canvasLocalBounds=TDRectangle();
        
        /* zoom to the old geographical bounds */
        ZoomToGeo(geoBounds);
        
        return oldproj;
}
        
/* Add a layer on top */
MapLayerInstance* MapEditor::PushLayer(MapLayer* layer) {
        MapLayerInstance* inst=new MapLayerInstance(layer);
        _mapLayerInstances.Append(inst);
        _mapLayerInstanceMap[layer]=inst;
        _mapLayerNameMap[layer->GetName()]=inst;
        
        layer->RegisterEditor(this);
        
        /* redraw the area affected by the given layer */
        wxRect layerRect=layer->GetDeviceBounds(this);
        DamageRect(layerRect);
        
        return inst;
}

/* Add a layer below the given layer */
MapLayerInstance* MapEditor::InsertLayer(MapLayer* layer, MapLayer* below) {
        MapLayerInstance* belowInstance=FindLayerInstanceByLayer(below);
        wxASSERT(belowInstance!=(MapLayerInstance*)0);
        MapLayerInstanceList::Node *belowNode=_mapLayerInstances.Find(belowInstance);
        wxASSERT(belowNode!=NULL);
        
        MapLayerInstance* inst=new MapLayerInstance(layer);
        _mapLayerInstances.Insert(belowNode,inst);
        _mapLayerInstanceMap[layer]=inst;
        _mapLayerNameMap[layer->GetName()]=inst;
        
        layer->RegisterEditor(this);
        
        /* redraw the area affected by the given layer */
        wxRect layerRect=layer->GetDeviceBounds(this);
        DamageRect(layerRect);
        
        return inst;
}

/* Find a layer instance by its name */
MapLayerInstance* MapEditor::FindLayerInstanceByName(wxString name) {
        MapLayerNameMap::iterator it=_mapLayerNameMap.find(name);
        if (it==_mapLayerNameMap.end())
                return (MapLayerInstance*)0;
        return it->second;
}

/* Find a layer instance by its layer */
MapLayerInstance* MapEditor::FindLayerInstanceByLayer(MapLayer* layer) {
        MapLayerInstanceMap::iterator it=_mapLayerInstanceMap.find(layer);
        if (it==_mapLayerInstanceMap.end())
                return (MapLayerInstance*)0;
        return it->second;
}

/* Remove a layer */
void MapEditor::RemoveLayer(MapLayer* layer) {
        MapLayerInstance* layerInstance=FindLayerInstanceByLayer(layer);
        wxASSERT(layerInstance!=(MapLayerInstance*)0);
        
        wxRect oldLayerRect=layer->GetDeviceBounds(this);
        /* redraw the area affected by the given layer */
        DamageRect(oldLayerRect);
        
        _mapLayerInstances.DeleteObject(layerInstance);
        _mapLayerNameMap.erase(layer->GetName());
        _mapLayerInstanceMap.erase(layer);
        
        layer->UnregisterEditor(this);
}

/* Add a pass on top */
MapPassInstance* MapEditor::PushPass(const MapPass& pass) {
        MapPassInstance* inst=new MapPassInstance(pass);
        _mapPassInstances.Append(inst);
        _mapPassInstanceMap[&pass]=inst;
        _mapPassNameMap[pass.GetName()]=inst;
        
        /*
         * Redraw the whole map (we don't know which parts are affected
         * by the pass)
         */
        DamageAll();
        
        return inst;
}

/* Add a pass below the given pass */
MapPassInstance* MapEditor::InsertPass(const MapPass& pass, const MapPass& below) {
        MapPassInstance* belowInstance=FindPassInstanceByPass(below);
        wxASSERT(belowInstance!=(MapPassInstance*)0);
        MapPassInstanceList::Node *belowNode=_mapPassInstances.Find(belowInstance);
        wxASSERT(belowNode!=NULL);
        
        MapPassInstance* inst=new MapPassInstance(pass);
        _mapPassInstances.Insert(belowNode,inst);
        _mapPassInstanceMap[&pass]=inst;
        _mapPassNameMap[pass.GetName()]=inst;
        
        /*
         * Redraw the whole map (we don't know which parts are affected
         * by the pass)
         */
        DamageAll();
        
        return inst;
}

/* Find a pass instance by its name */
MapPassInstance* MapEditor::FindPassInstanceByName(wxString name) {
        MapPassNameMap::iterator it=_mapPassNameMap.find(name);
        if (it==_mapPassNameMap.end())
                return (MapPassInstance*)0;
        return it->second;
}

/* Find a pass instance by its pass */
MapPassInstance* MapEditor::FindPassInstanceByPass(const MapPass& pass) {
        MapPassInstanceMap::iterator it=_mapPassInstanceMap.find(&pass);
        if (it==_mapPassInstanceMap.end())
                return (MapPassInstance*)0;
        return it->second;
}

/* Remove a pass */
void MapEditor::RemovePass(const MapPass& pass) {
        MapPassInstanceMap::iterator it=_mapPassInstanceMap.find(&pass);
        MapPassInstance* passInstance=it->second;
        wxASSERT(passInstance!=(MapPassInstance*)0);
        
        _mapPassInstances.DeleteObject(passInstance);
        _mapPassNameMap.erase(pass.GetName());
        _mapPassInstanceMap.erase(it);
        
        /*
         * Redraw the whole map (we don't know which parts were affected
         * by the pass)
         */
        DamageAll();
}

wxMenuItem* MapEditor::GetMenuItemFromCommandEvent(wxCommandEvent& evt) {
        wxMenu* menu=wxDynamicCast(evt.GetEventObject(),wxMenu);
	if (menu) {
		return menu->FindItem(evt.GetId());
	}

	wxFrame* frame=wxDynamicCast(evt.GetEventObject(),wxFrame);
	wxASSERT(frame);
	wxMenuBar* menubar=frame->GetMenuBar();
	wxASSERT(menubar);
	return menubar->FindItem(evt.GetId());
}

void MapEditor::OnToggleLayerVisibility(wxCommandEvent& evt) {
        wxMenuItem* menuItem=GetMenuItemFromCommandEvent(evt);
        const wxString& layerName=menuItem->GetLabel();
        MapLayerInstance* layerInstance=FindLayerInstanceByName(layerName);
        if (evt.GetInt()) {
                layerInstance->Enable();
                
        } else {
                layerInstance->Disable();
        }

        wxRect oldLayerRect=layerInstance->GetLayer()->GetDeviceBounds(this);
        DamageRect(oldLayerRect);
}

void MapEditor::OnToggleLayerLock(wxCommandEvent& evt) {
        wxMenuItem* menuItem=GetMenuItemFromCommandEvent(evt);
        const wxString& layerName=menuItem->GetLabel();
        MapLayerInstance* layerInstance=FindLayerInstanceByName(layerName);
        if (evt.GetInt()) {
                layerInstance->Lock();
        } else {
                layerInstance->Unlock();
        }
}

void MapEditor::OnTogglePassVisibility(wxCommandEvent& evt) {
        wxMenuItem* menuItem=GetMenuItemFromCommandEvent(evt);
        const wxString& passName=menuItem->GetLabel();
        MapPassInstance* passInstance=FindPassInstanceByName(passName);
        if (evt.GetInt()) {
                passInstance->Enable();
        } else {
                passInstance->Disable();
        }
        /*
         * Redraw the whole map (we don't know which parts were affected
         * by the pass)
         */
        DamageAll();
}

void MapEditor::DamageRect(wxRect clientRect) {
        wxASSERT(_drawingWindow!=NULL);
        _drawingWindow->Refresh(FALSE,&clientRect);
        AdjustScrollbars();
}

void MapEditor::DamageAll() {
        if (_drawingWindow)
                _drawingWindow->Refresh();
        AdjustScrollbars();
}

void MapEditor::Scroll(wxSize devDelta, bool updateCanvasBounds) {
        Point3D delta=GetCoordinateSystem().DeviceToLogicalRel(devDelta);
        
        double ox,oy;
        GetCoordinateSystem().GetLogicalOrigin(ox,oy);
        ox+=delta.x();
        oy+=delta.y();
        GetCoordinateSystem().SetLogicalOrigin(ox,oy);
        GetDrawingWindow()->ScrollWindow(-devDelta.GetWidth(),-devDelta.GetHeight());

        MapEditorDecorationList::Node *node;
        
        wxSize invDelta=wxSize(-devDelta.GetWidth(),-devDelta.GetHeight());
        
        for (node=_backgroundDecos.GetFirst();node;node=node->GetNext())
                node->GetData()->Scroll(invDelta);
        for (node=_foregroundDecos.GetFirst();node;node=node->GetNext())
                node->GetData()->Scroll(invDelta);
        if (updateCanvasBounds) {
                _canvasLocalBounds=GetVisibleLocalBounds();
        }
        AdjustScrollbars();
}

void MapEditor::Scroll(float deltaFractionX, float deltaFractionY) {
        wxSize devSize=GetDrawingWindow()->GetClientSize();
        devSize.SetWidth((int)(devSize.GetWidth()*deltaFractionX));
        devSize.SetHeight((int)(devSize.GetHeight()*deltaFractionY));
        Scroll(devSize);
}

void MapEditor::Zoom(Point3D localCenter, double factor) {
        double scale=GetCoordinateSystem().GetLogicalScale();
        double ox,oy;
        GetCoordinateSystem().GetLogicalOrigin(ox,oy);
        ox=(ox-localCenter.x())/factor+localCenter.x();
        oy=(oy-localCenter.y())/factor+localCenter.y();
        GetCoordinateSystem().SetLogicalScale(scale*factor);
        GetCoordinateSystem().SetLogicalOrigin(ox,oy);
        DamageAll();
}

void MapEditor::Zoom(double factor) {
        double scale;
        scale=GetCoordinateSystem().GetLogicalScale()*factor;
        GetCoordinateSystem().SetLogicalScale(scale);
        _canvasLocalBounds=GetVisibleLocalBounds();
        DamageAll();
}

void MapEditor::SetZoom(double zoom) {
        GetCoordinateSystem().SetLogicalScale(zoom);
        _canvasLocalBounds=GetVisibleLocalBounds();
        DamageAll();
}

void MapEditor::ZoomTo(TDRectangle targetBounds) {
        if (targetBounds.GetWidth()<=0 || targetBounds.GetHeight()<=0)
                return;
        if (GetDrawingWindow()==NULL)
                return;
        
        int cw,ch;
        GetDrawingWindow()->GetClientSize(&cw,&ch);
        double sx,sy,scale;
        Point3D center=targetBounds.GetCenter();
        
        sx=cw/targetBounds.GetWidth();
        sy=ch/targetBounds.GetHeight();
        scale=std::min(sx,sy);
        GetCoordinateSystem().SetLogicalScale(scale);
        GetCoordinateSystem().SetLogicalOrigin(center.x(),center.y());
        _canvasLocalBounds=GetVisibleLocalBounds();
        DamageAll();
}

void MapEditor::ZoomToGeo(TDRectangle geoBounds) {
        MapProjection* proj=GetProjection();
        Point3D nw,ne,se,sw;
        
        nw=proj->ConvertToLocal(geoBounds.GetTopLeft());
        ne=proj->ConvertToLocal(geoBounds.GetTopRight());
        se=proj->ConvertToLocal(geoBounds.GetBottomRight());
        sw=proj->ConvertToLocal(geoBounds.GetBottomLeft());
        
        TDRectangle bounds=TDRectangle::Union(TDRectangle(nw,se),TDRectangle(ne,sw));
        
        ZoomTo(bounds);
}

void MapEditor::AdjustScrollbars() {
        TDRectangle scrollBounds;
        TDRectangle visibleBounds=GetVisibleLocalBounds();
        
        scrollBounds=GetLocalBounds();
        scrollBounds.Scale(1.5,1.5);
        
        scrollBounds=TDRectangle::Union(scrollBounds,visibleBounds);
        
        /* ensure that the canvas bounds are always included */
        scrollBounds=TDRectangle::Union(scrollBounds,_canvasLocalBounds);
        
        MapCoordinateSystem& cs=GetCoordinateSystem();
        wxRect visibleBoundsDev=cs.LogicalToDevice(visibleBounds);
        wxRect scrollBoundsDev=cs.LogicalToDevice(scrollBounds);
        
        int dx=(int)((visibleBoundsDev.GetLeft()-scrollBoundsDev.GetLeft())/Scroll_Line_Width+0.5);
        int dy=(int)((visibleBoundsDev.GetTop()-scrollBoundsDev.GetTop())/Scroll_Line_Height+0.5);
        int w=(int)(visibleBoundsDev.GetWidth()/Scroll_Line_Width+0.5);
        int h=(int)(visibleBoundsDev.GetHeight()/Scroll_Line_Height+0.5);
        int totalw=(int)(scrollBoundsDev.GetWidth()/Scroll_Line_Width+0.5);
        int totalh=(int)(scrollBoundsDev.GetHeight()/Scroll_Line_Height+0.5);
        
        _drawingWindow->SetScrollbar(wxHORIZONTAL,dx,w,totalw);
        _drawingWindow->SetScrollbar(wxVERTICAL,dy,h,totalh);
}

void MapEditor::OnScrollWin(wxScrollWinEvent& event) {
        TDRectangle scrollBounds;
        TDRectangle visibleBounds=GetVisibleLocalBounds();
        
        scrollBounds=GetLocalBounds();
        scrollBounds.Scale(1.5,1.5);
        
        scrollBounds=TDRectangle::Union(scrollBounds,visibleBounds);
        
        /* ensure that the canvas bounds are always included */
        scrollBounds=TDRectangle::Union(scrollBounds,_canvasLocalBounds);
        
        MapCoordinateSystem& cs=GetCoordinateSystem();
        wxPoint scrollTopLeft=cs.LogicalToDevice(scrollBounds.GetTopLeft());
        wxPoint oldTopLeft=cs.LogicalToDevice(visibleBounds.GetTopLeft());
        wxPoint newTopLeft=oldTopLeft;
        int pos=event.GetPosition();
        switch (event.GetOrientation()) {
        case wxHORIZONTAL:
                newTopLeft.x=scrollTopLeft.x+pos*Scroll_Line_Width;
                break;
        case wxVERTICAL:
                newTopLeft.y=scrollTopLeft.y+pos*Scroll_Line_Height;
                break;
        }
        wxSize devDelta=wxSize(newTopLeft.x-oldTopLeft.x,newTopLeft.y-oldTopLeft.y);
        Scroll(devDelta,false);
}

TDRectangle MapEditor::GetLocalBounds() {
        TDRectangle r;
        MapLayerInstanceList::Node *layerNode;
        for (layerNode=_mapLayerInstances.GetFirst();
             layerNode;
             layerNode=layerNode->GetNext()) {
		if (layerNode->GetData()->IsEnabled()) {
			TDRectangle layerBounds;
			layerBounds=layerNode->GetData()->GetLayer()->GetLocalBounds(this);
                        r=TDRectangle::Union(r,layerBounds);
		}
        }
        r=TDRectangle::Union(r,GetSelectionManager().GetLocalBounds());
        return r;
}

TDRectangle MapEditor::GetGeographicBounds() {
        TDRectangle r;
        MapLayerInstanceList::Node *layerNode;
        for (layerNode=_mapLayerInstances.GetFirst();
             layerNode;
             layerNode=layerNode->GetNext()) {
		if (layerNode->GetData()->IsEnabled()) {
			TDRectangle layerBounds;
			layerBounds=layerNode->GetData()->GetLayer()->GetGeographicBounds();
                        r=TDRectangle::Union(r,layerBounds);
		}
        }
        return r;
}

TDRectangle MapEditor::GetVisibleLocalBounds() {
        TDRectangle logicalBounds;
        
        int cw,ch;
        GetDrawingWindow()->GetClientSize(&cw,&ch);
        MapCoordinateSystem& cs=GetCoordinateSystem();
        
        Point3D topLeft,bottomRight;
        topLeft=cs.DeviceToLogical(wxPoint(0,0));
        bottomRight=cs.DeviceToLogical(wxPoint(cw,ch));
        
        return TDRectangle(topLeft,bottomRight);
}

TDRectangle MapEditor::GetVisibleGeographicBounds() {
        TDRectangle logicalBounds=GetVisibleLocalBounds();
        
        MapProjection* proj=GetProjection();
        
        Point3D nw,ne,se,sw;
        
        nw=proj->ConvertFromLocal(logicalBounds.GetTopLeft());
        ne=proj->ConvertFromLocal(logicalBounds.GetTopRight());
        se=proj->ConvertFromLocal(logicalBounds.GetBottomRight());
        sw=proj->ConvertFromLocal(logicalBounds.GetBottomLeft());
        
        return TDRectangle::Union(TDRectangle(nw,se),TDRectangle(ne,sw));
}

void MapEditor::GetObjectsOfClass(MapObjectList& list, wxClassInfo* targetClass) {
        MapLayerInstanceList::Node *layerNode;
        for (layerNode=_mapLayerInstances.GetFirst();
             layerNode;
             layerNode=layerNode->GetNext()) {
                layerNode->GetData()->GetLayer()->GetObjectsOfClass(list,targetClass);
        }
}

void MapEditor::GetHitObjects(MapObjectList& list, TDRectangle r, bool includeIntersected) {
        MapLayerInstanceList::Node *layerNode;
        for (layerNode=_mapLayerInstances.GetFirst();
             layerNode;
             layerNode=layerNode->GetNext()) {
               if (layerNode->GetData()->IsEnabled() &&
                   !layerNode->GetData()->IsLocked())
                        layerNode->GetData()->GetLayer()->GetHitObjects(list,this,r,includeIntersected);
        }
}

void MapEditor::ProvideActionMenu(wxMenu& menu) {
        wxMenu *showLayersMenu,*lockLayersMenu,*showPassesMenu;
        showLayersMenu=new wxMenu();
        lockLayersMenu=new wxMenu();
        showPassesMenu=new wxMenu();
        
        showLayersMenu->SetEventHandler(menu.GetEventHandler());
        lockLayersMenu->SetEventHandler(menu.GetEventHandler());
        showPassesMenu->SetEventHandler(menu.GetEventHandler());
        
        MapLayerInstanceList::Node *layerNode;
        for (layerNode=_mapLayerInstances.GetFirst();
             layerNode;
             layerNode=layerNode->GetNext()) {
                wxMenuItem *item;
                
                item=showLayersMenu->AppendCheckItem(wxID_ANY,layerNode->GetData()->GetName());
                menu.GetEventHandler()->Connect(item->GetId(),
                                                wxEVT_COMMAND_MENU_SELECTED,
                                                wxCommandEventHandler(MapEditor::OnToggleLayerVisibility),
                                                NULL,
                                                this);
                item->Check(layerNode->GetData()->IsEnabled());
                item=lockLayersMenu->AppendCheckItem(wxID_ANY,layerNode->GetData()->GetName());
                menu.GetEventHandler()->Connect(item->GetId(),
                                                wxEVT_COMMAND_MENU_SELECTED,
                                                wxCommandEventHandler(MapEditor::OnToggleLayerLock),
                                                NULL,
                                                this);
                item->Check(layerNode->GetData()->IsLocked());
        }
        
        MapPassInstanceList::Node *passNode;
        for (passNode=_mapPassInstances.GetFirst();
             passNode;
             passNode=passNode->GetNext()) {
                wxMenuItem *item;
                
                item=showPassesMenu->AppendCheckItem(wxID_ANY,passNode->GetData()->GetName());
                menu.GetEventHandler()->Connect(item->GetId(),
                                                wxEVT_COMMAND_MENU_SELECTED,
                                                wxCommandEventHandler(MapEditor::OnTogglePassVisibility),
                                                NULL,
                                                this);
                item->Check(passNode->GetData()->IsEnabled());
        }
             
        menu.Append(0,wxT("Show Layers"),showLayersMenu);
        menu.Append(0,wxT("Lock Layers"),lockLayersMenu);
        menu.Append(0,wxT("Show Passes"),showPassesMenu);
}

void MapEditor::OnPaint(wxPaintEvent& evt) {
        wxPaintDC dc(GetDrawingWindow());
        
        /* Prepare a wxMemoryDC to avoid flickering */
        wxCoord x,y,w,h;
        dc.GetClippingBox(&x,&y,&w,&h);
        if (w==0 && h==0)
                GetDrawingWindow()->GetClientSize(&w,&h);
        wxBitmap bitmap=wxBitmap(w,h);

        wxMemoryDC mdc;
        mdc.SetFont(dc.GetFont());
        mdc.SetDeviceOrigin(x,y);

        mdc.SelectObject(bitmap);
        mdc.SetClippingRegion(x,y,w,h);

        mdc.SetBackground(wxBrush(Map_Background_Colour));
        mdc.Clear();

        /* Do the actual painting */
        MapEditorDecorationList::Node *node;
        MapLayerInstanceList::Node *layerNode;
        MapPassInstanceList::Node* passnode;
        
        /* Paint background decorations */
        for (node=_backgroundDecos.GetFirst();node;node=node->GetNext())
                node->GetData()->Paint(mdc);
        
        /* Paint the map, possibly in multiple passes */
        for (passnode=_mapPassInstances.GetFirst();
             passnode;
             passnode=passnode->GetNext()) {
                if (!passnode->GetData()->IsEnabled())
                        continue; /* skip disabled passes */
                
                const MapPass& pass=passnode->GetData()->GetPass();
                
                for (layerNode=_mapLayerInstances.GetFirst();
                     layerNode;
                     layerNode=layerNode->GetNext()) {
                        if (!layerNode->GetData()->IsEnabled())
                                continue;
                        layerNode->GetData()->GetLayer()->Paint(mdc,this,pass);
                }
        }
        
        /* Paint foreground decorations */
        for (node=_foregroundDecos.GetFirst();node;node=node->GetNext())
                node->GetData()->Paint(mdc);
        
        /* Paint the selections */
        GetSelectionManager().Paint(mdc);
        
        /* Copy the memory DC to the window */
        dc.Blit(x,y,w,h,&mdc,x,y);
        
        /* That's it! Pass event on */
        evt.Skip();
}

void MapEditor::OnResize(wxSizeEvent& evt) {
        wxCoord vw,vh;
        GetDrawingWindow()->GetClientSize(&vw,&vh);
        GetCoordinateSystem().SetDeviceOrigin(vw/2,vh/2);
        AdjustScrollbars();
        GetDrawingWindow()->Refresh();
}

void MapEditor::OnMouseEnter(wxMouseEvent& evt) {
        GetModeManager().UpdateStatus();
}

