#ifndef MapLayer_h
#define MapLayer_h

// For compilers that support precompilation, includes "wx.h".
#include <wx/wxprec.h>

#ifdef __BORLANDC__
#pragma hdrstop
#endif

#ifndef WX_PRECOMP
// Include your minimal set of headers here, or wx.h
#include <wx/wx.h>
#endif

#include "MapObject.h"
#include <Transactions/TransactionManager.h>
#include <Transactions/TransactionObject.h>

class MapEditor;
class MapPass;

WX_DECLARE_LIST(MapEditor,MapLayerEditors);

/**
 * \brief A MapLayer represents a layer of a map.
 *
 * Map layers can be stacked on the display.
 */
class MapLayer : public TransactionObject {
        
        DECLARE_ABSTRACT_CLASS(MapLayer)
        TRANSACTIONOBJECT_DEFINE_SMARTREF(MapLayer);
public:
	MapLayer(TransactionManager* tmgr, wxString name);
        virtual ~MapLayer();
        
        /**
         * \brief Paint this layer.
         *
         * \param dc            The device context to draw on.
         * \param editor        The editor providing further context.
         * \param pass          The pass to draw.
         */
        virtual void Paint(wxDC& dc, MapEditor* editor, const MapPass& pass)=0;
        
        /**
         * \brief Collect all objects in this layer contained or intersected by
         * a given rectangle.
         *
         * \param list          The list to add to.
         * \param editor        The editor in whose context the rectangle is
         *                      defined.
         * \param r             The hit-rectangle.
         *
         * \todo Design: This is actually bad design as this really only
         *       provided by MapObjectContainer.
         */
        virtual void GetHitObjects(MapObjectList& list,
                                   MapEditor* editor,
                                   TDRectangle r,
                                   bool includeIntersected=true)=0;
        
        /**
         * \brief Schedule this layer for repaint.
         */
        virtual void Damage();
        
        /**
         * \brief Return the bounds of the layer in device coordinates.
         *
         * \param editor        The editor defining the logical coordinate system.
         */
        virtual wxRect GetDeviceBounds(MapEditor* editor)=0;
         
        /**
         * \brief Return the bounds of the layer in logical coordinates.
         *
         * \param editor        The editor defining the logical coordinate system.
         */
        virtual TDRectangle GetLocalBounds(MapEditor* editor)=0;

        /**
         * \brief Return the geographic bounds of the layer.
         */
        virtual TDRectangle GetGeographicBounds()=0;
        
        /**
         * \brief Get a list of objects of the given class from this lay.er
         *
         * \todo Design: This is actually bad design as this really only
         *       provided by MapObjectContainer.
         */
        virtual void GetObjectsOfClass(MapObjectList& list, wxClassInfo* targetClass)=0;
        
        /**
         * \brief Register this layer as being displayed in the given editor.
         */
        void RegisterEditor(MapEditor* editor);

        /**
         * \brief Unregister this layer with the given editor.
         */
        void UnregisterEditor(MapEditor* editor);
        
        /**
         * \brief Get the list of editors this layer is displayed in.
         */
        const MapLayerEditors& GetEditors() const { return _editors; }
        
        /**
         * \brief Get the name of this layer.
         */
        wxString GetName() const { return _name; }
protected:
        wxString _name;
        MapLayerEditors _editors;
};
#endif

