#ifndef MapPropertyManager_h
#define MapPropertyManager_h

// For compilers that support precompilation, includes "wx.h".
#include <wx/wxprec.h>

#ifdef __BORLANDC__
#pragma hdrstop
#endif

#ifndef WX_PRECOMP
// Include your minimal set of headers here, or wx.h
#include <wx/wx.h>
#endif

#include "MapSelectionChangeListener.h"

#include <Transactions/TransactionObject.h>
#include <Transactions/TransactionObjectChangeListener.h>

class MapEditor;
class wxPropertyGrid;
class wxPropertyGridEvent;

class MapUpdatePropertiesEvent;

DECLARE_EVENT_TYPE(wxEVT_MAP_UPDATE_PROPERTIES,-1);

/**
 * \brief A MapPropertyManager manages a property grid, handling property
 * changes and updates.
 *
 * There can be multiple property managers for each MapEditor, each managing
 * their own property grid.
 *
 * \note The property manager must be destroyed manually before destroying the
 * associated MapEditor object.
 */
class MapPropertyManager: public wxEvtHandler,
                          public TransactionObjectChangeListener {
        
        DECLARE_CLASS(MapPropertyManager)
        
public:
	MapPropertyManager(wxPropertyGrid* grid);
        virtual ~MapPropertyManager();
        
        /**
         * \brief Schedule a property update.
         *
         * Requests for property updates, e.g., due to changes on objects or
         * on the list of selections, are collected and postponed similar to
         * paint events.
         *
         * This method checks whether an update is already scheduled and if not,
         * posts an update event to this object.
         *
         * \note Rationale: Often multiple changes to an object happen in sequence,
         * e.g., when restoring an object from a memento. In that case, the
         * restore method from the superclass is called first, triggering an
         * object change message. This leads to multiple object change messages
         * being sent.
         *
         * \par
         * In order to avoid multiple costly property updates, the updates are
         * postponed. The scheduled update event is guaranteed to be delivered
         * after any yet unprocessed events which might possibly result in
         * additional changed.
         */
        void SchedulePropertyUpdate();
        
        void OnUpdateProperties(MapUpdatePropertiesEvent&);
        
        /**
         * \brief Update the property grid.
         */
        virtual void UpdateProperties();
        
        /**
         * \brief React on a property change.
         *
         * This is the event handler for the EVT_PG_CHANGED event from the
         * property grid.
         */
        virtual void PropertyChanged(wxPropertyGridEvent& evt)=0;

        virtual void BeforeObjectChange(TransactionObject* object);

        /**
         * \brief React on a change of a selected object.
         *
         * This simply schedules a property update event using SchedulePropertyUpdate().
         */
        virtual void AfterObjectChange(TransactionObject* object);
        
        wxPropertyGrid* GetManagedGrid() const { return _managedGrid; }
protected:
        wxPropertyGrid* _managedGrid;
        bool _updateEventPosted;

        DECLARE_EVENT_TABLE()
};

class MapUpdatePropertiesEvent: public wxEvent {
        
        DECLARE_CLASS(MapUpdatePropertiesEvent)
        
public:
	MapUpdatePropertiesEvent(wxEventType eventType):
                wxEvent(0,eventType) {}
        
        MapUpdatePropertiesEvent(const MapUpdatePropertiesEvent& evt):
                wxEvent(evt) {}
        
        virtual wxEvent* Clone() const { return new MapUpdatePropertiesEvent(*this); }
};


typedef void (wxEvtHandler::*MapUpdatePropertyEventFunction)(MapUpdatePropertiesEvent&);

#define EVT_MAP_UPDATE_PROPERTIES(fn) \
    DECLARE_EVENT_TABLE_ENTRY( wxEVT_MAP_UPDATE_PROPERTIES, wxID_ANY, -1, \
    (wxObjectEventFunction) (wxEventFunction) wxStaticCastEvent(MapUpdatePropertyEventFunction,&fn) ,\
    (wxObject *) NULL ),

#endif

