#ifndef ModeManager_h
#define ModeManager_h

// For compilers that support precompilation, includes "wx.h".
#include <wx/wxprec.h>

#ifdef __BORLANDC__
#pragma hdrstop
#endif

#ifndef WX_PRECOMP
// Include your minimal set of headers here, or wx.h
#include <wx/wx.h>
#endif

class Mode;
class MapEditor;

WX_DECLARE_LIST(Mode,ModeList);

/**
 * \brief A ModeManager manages a stack of Mode instances.

 * All events received by the ModeManager traverse
 * the mode-stack from top to bottom, terminating as soon
 * as a mode has handled the event or the bottom of the stack
 * is reached.
 */
class ModeManager : public wxEvtHandler {
        DECLARE_CLASS(ModeManager)
public:
	ModeManager(MapEditor* editor);
	~ModeManager();
	
        /**
         * \brief Get the wxStatusBar on which the status provided by the modes
         * is displayed.
         */
	wxStatusBar* GetStatusBar();
        
        /**
         * \brief Get the number of the field of the wxStatusBar in which
         * the Mode-provided status is displayed.
         */
	int GetStatusField() { return _statusField; }

        /**
         * \brief Set the number of the field of the wxStatusBar in which
         * the Mode-provided status is displayed.
         */
	void SetStatusField(int field) { _statusField=field; }
	
	/**
         * \brief Update the status bar.
         *
         * \todo Ignore empty status strings from modes.
         */
	void UpdateStatus();
        
	/**
         * \brief Insert the given mode on top of the stack.
         */
	void InsertOnTop(Mode* mode);
        
        /**
         * \brief Insert the given mode above the given other mode.
         *
         * \param mode  The mode to insert
         * \param above The mode above which to insert
         */
        void InsertAbove(Mode* mode, Mode* above);
        
        
        /**
         * \brief Insert the given mode below the given other mode.
         *
         * \param mode  The mode to insert
         * \param below The mode below which to insert
         */
        void InsertBelow(Mode* mode, Mode* below);
        
	/**
         * \brief Insert the given mode at the bottom of the stack.
         */
        void InsertAtBottom(Mode* mode);
        
        /**
         * \brief Remove the topmost mode.
         *
         * \param force         Flag indicating whether to override a vetor
         *                      by Mode::CanExit()
         *
         * \return false if the stack is empty or the mode vetoed its removal
         *               and force was false.
         */
        bool RemoveTopmost(bool force=false);
        
        /**
         * \brief Remove the given mode from the stack.
         *
         * \param mode          The mode to remove
         * \param force         Flag indicating whether to override a vetor
         *                      by Mode::CanExit()
         *
         * \return false if the stack does not contain the mode or the mode
         *               vetoed its removal and force was false.
         */
        bool Remove(Mode* mode, bool force=false);
        
        /**
         * \brief Remove all modes from the stack.
         *
         * This method repeatedly calls RemoveTopmost with force as parameter
         * until RemoveTopmost returns false.
         *
         * \param force         Flag indicating whether to override a vetor
         *                      by Mode::CanExit()
         */
        void RemoveAll(bool force=false);
        
        /**
         * \brief Remove all modes starting from a given mode.
         *
         * This method repeatedly calls RemoveTopmost with force as parameter
         * until startMode was removed.
         *
         * \param startMode     The mode from which to start removing upwards.
         * \param force         Flag indicating whether to override a vetor
         *                      by Mode::CanExit()
         *
         * \return false if not all modes could be removed or the given startMode
         *         is not on the stack, true otherwise.
         */
        bool RemoveAt(Mode* startMode, bool force=false);
        
        virtual bool ProcessEvent(wxEvent& evt);
        
        MapEditor* GetEditor() const { return _editor; }
protected:
        /* Initialise the mode properly */
        void SetupMode(Mode* mode);
        
	ModeList _modes;
        
	int _statusField;
        MapEditor* _editor;
};

#endif
