#include "CollisionUtils.h"
#include "point3d.h"

#include <iostream>

bool CollisionUtils::ConvexPolygonsCollide(size_t n1, const Point3D* pts1,
                                           size_t n2, const Point3D* pts2)
{
        /*
         * The algorithm is based on the Separating Axis Theorem.
         * (c.f. http://en.wikipedia.org/wiki/Separating_axis_theorem)
         *
         * The theorem says that two objects overlap if and only if there is
         * no line which can be put between them.
         *
         * For two convex polygons the intersection test therefore can be
         * done by checking whether the projection of the polygons along the
         * normals of each of their sides overlap.
         */
         
         return (CheckOverlap(n1,pts1,n2,pts2) && CheckOverlap(n2,pts2,n1,pts1));
}

bool CollisionUtils::CheckOverlap(size_t n1, const Point3D* pts1,
                                  size_t n2, const Point3D* pts2)
{
         for (size_t i=0;i<n1;i++) {
                 const Point3D &p0=pts1[i];
                 const Point3D &p1=pts1[(i+1)%n1];
                 
                 double l=p1.distance3D(p0);
                 Point3D v=p1-p0;
                 Point3D n=Point3D(-v.y()/l,v.x()/l,0);
                 
                 double min1,max1;
                 min1=max1=pts1[0].x()*n.x()+pts1[0].y()*n.y();
                 
                 for (size_t j=1;j<n1;j++) {
                         double v=pts1[j].x()*n.x()+pts1[j].y()*n.y();
                         if (v<min1)
                                 min1=v;
                         if (v>max1)
                                 max1=v;
                 }
                 
                 double min2,max2;
                 min2=max2=pts2[0].x()*n.x()+pts2[0].y()*n.y();

                 for (size_t j=1;j<n2;j++) {
                         double v=pts2[j].x()*n.x()+pts2[j].y()*n.y();
                         if (v<min2)
                                 min2=v;
                         if (v>max2)
                                 max2=v;
                 }
                 
                 if (min1>max2 || min2>max1) {
                         /* no overlap */
                         return false;
                 }
         }
         return true;
}

bool CollisionUtils::ConvexPolygonRectangleCollide(TDRectangle r, size_t n, const Point3D *pts) {
        Point3D rpts[4];
        
        rpts[0]=r.GetTopLeft();
        rpts[1]=r.GetTopRight();
        rpts[2]=r.GetBottomRight();
        rpts[3]=r.GetBottomLeft();
        
        return ConvexPolygonsCollide(4,rpts,n,pts);
}

bool CollisionUtils::IntersectsCircle(TDRectangle rect, Point3D c, double r) {
        TDRectangle bounds=TDRectangle(c,2*r,2*r);
        if (!rect.Intersects(bounds))
                return false;
        
        double xmin,xmax;
        double ymin,ymax;
        
        rect.Translate(-c.x(),-c.y());
        double x1,x2,y1,y2;
        x1=rect.GetLeft();
        x2=rect.GetRight();
        y1=rect.GetBottom();
        y2=rect.GetTop();
        
        xmin=std::max(-r,x1);
        xmax=std::min(r,x2);
        
        if (xmin>xmax)
                return false;
        
        ymin=std::max(-r,y1);
        ymax=std::min(r,y2);
        
        if (xmin>0) {
                double s=sqrt(r*r-xmin*xmin);
                ymin=std::max(ymin,-s);
                ymax=std::min(ymax,s);
        }
        
        if (xmax<0) {
                double s=sqrt(r*r-xmax*xmax);
                ymin=std::max(ymin,-s);
                ymax=std::min(ymax,s);
        }
        
        return (ymin<=ymax);
}

